import urllib
import asyncio
import logging
from typing import Set, Optional
from urllib.parse import urlparse

import aiohttp
from bs4 import BeautifulSoup

from .config import Config

logger = logging.getLogger(__name__)

__all__ = ("WebInstrument",)


class WebInstrument:
    def __init__(self, init_url: str, config: Config):
        """
        Core class for working with webpages:

        1. get webpages

        2. get webpage tags

        3. filter links

        Args:
            init_url: webpage main link
            config: dataclass contains all params
        """
        self.init_url = init_url
        self.domain = self.get_domain(url=self.init_url)
        self.config = config

    @staticmethod
    def get_domain(url: str) -> str:
        """
        Method parse link to get core domain from it
        Args:
            url: webpage link

        Returns:
            domain name
        """
        return ".".join(urlparse(url=url).hostname.split(".")[-2:])

    @staticmethod
    def find_tags(page_data: str, tag: str, key: str) -> Set[str]:
        """
        Parses webpage text and extracts values of a specific attribute from given tags,
        ignoring elements with rel="nofollow".

        Args:
            page_data: HTML content of the page
            tag: HTML tag to search for (e.g., 'a')
            key: attribute to extract (e.g., 'href')

        Returns:
            Set of extracted attribute values
        """
        result_elements = set()

        soup = BeautifulSoup(page_data, "html.parser")
        elements = soup.find_all(tag)

        for element in elements:
            if not element.has_attr(key):
                continue

            if rel_values := element.get("rel"):
                if "nofollow" in rel_values:
                    continue

            if value := element.get(key).strip():
                result_elements.add(value)

        return result_elements

    async def download_page(self, url: str) -> Optional[str]:
        """
        Method connect open webpage and download it's as text
        Args:
            url: webpage for downloading

        Returns:
            Webpage as text
        """
        async with aiohttp.ClientSession(headers=self.config.header) as session:
            for attempt in self.attempts_generator():
                try:
                    async with session.get(url=url) as resp:
                        if resp.status == 200:
                            logger.info(f"Page success loaded - {url = }")
                            return await resp.text()
                        else:
                            await asyncio.sleep(1 * attempt)
                            raise ValueError(
                                f"Wrong response status {attempt = }, {url = } ; {resp.status = }, {await resp.text()}"
                            )
                except Exception as err:
                    logger.warning(f"{err}")
            else:
                logger.error(f"Page not loaded - {url = }")

    @staticmethod
    def __filter_links_query(links: Set[str], is_query_enabled: bool = True) -> Set[str]:
        """
        Method filter webpages links set and return only links with same domain or subdomain
        Args:
            links: set of links for filtering
            is_query_enabled: accept or not links with query strings

        Returns:
            Filtered list of links
        """
        result_links = set()
        for link in links:
            if is_query_enabled and urlparse(url=link).query:
                result_links.add(link)
            elif not urlparse(url=link).query:
                result_links.add(link)
        return result_links

    def filter_links_domain(self, links: Set[str], is_subdomain: bool = True) -> Set[str]:
        """
        Method filter webpages links set and return only links with same domain or subdomain
        Args:
            links: set of links for filtering
            is_subdomain: accept or not links with subdomain

        Returns:
            Filtered list of links
        """
        result_links = set()
        check_logic = "endswith" if is_subdomain else "__eq__"
        for link in links:
            link_domain = urlparse(url=link).hostname
            if link_domain and getattr(link_domain, check_logic)(self.domain):
                result_links.add(link)
        return result_links

    @staticmethod
    def filter_inner_links(links: Set[str]) -> Set[str]:
        """
        Method get set of links and filter them from non-inner website links
        Args:
            links: set of website links

        Returns:
            Filtered list of links only with inner links
        """
        result_links = set()
        for link in links:
            if link and not link.startswith("https://"):
                result_links.add(link)
        return result_links

    def filter_links(self, canonical_url: str, links: Set[str]) -> Set[str]:
        filtered_links = set()
        # filter only local weblinks
        inner_links = self.filter_inner_links(links=links)
        # filter global domain weblinks from local links
        filtered_links.update(
            self.filter_links_domain(
                links=links.difference(inner_links),
                is_subdomain=self.config.accept_subdomains,
            )
        )
        # create fixed inner links (fixed - added to local link page url)
        filtered_links.update({urllib.parse.urljoin(canonical_url, inner_link) for inner_link in inner_links})
        # filter weblinks from webpages link minus links with query
        return self.__filter_links_query(links=filtered_links, is_query_enabled=self.config.is_query_enabled)

    @staticmethod
    def attempts_generator(amount: int = 6) -> int:
        """
        Function generates a generator of length equal to `amount`

        Args:
            amount: number of attempts generated

        Returns:
            Attempt number
        """
        yield from range(1, amount)



import logging
from typing import Set, List

from .instruments import WebInstrument, FileInstrument
from .instruments.config import Config

logger = logging.getLogger(__name__)
__all__ = ("LinksCrawler",)


class LinksCrawler:
    def __init__(self, init_url: str, config: Config):
        self.config = config
        self.web_instrument = WebInstrument(init_url=init_url, config=self.config)
        self.crawled_links: List[str]
        self.file_instrument = FileInstrument(file_name=self.config.file_name)

    async def __links_crawler(self, url: str, current_depth: int = 0) -> Set[str]:
        """
        Method with recursion for webpages crawling
        Args:
            url: url for read and parse weblinks
            current_depth: current recursion depth
        Returns:
            Set of weblinks from page
        """
        logger.info(f"Crawling page - {url} , depth - {current_depth}")
        if current_depth >= self.config.max_depth:
            return set()

        links = set()
        if page_data := await self.web_instrument.download_page(url=url):
            page_links = self.web_instrument.find_tags(page_data=page_data, tag="a", key="href")
            links = self.web_instrument.filter_links(canonical_url=url, links=page_links)
            rec_parsed_links = set()
            for link in sorted(links, key=len):
                rec_parsed_links.update(await self.__links_crawler(url=link, current_depth=current_depth + 1))

            links.update(rec_parsed_links)
        return links

    async def run(self) -> "LinksCrawler":
        """
        Method runs website crawling process
        Returns:
            Set with all crawled website pages links
        """
        logger.info(f"Starting crawling - {self.web_instrument.init_url}," f" config - {self.config}")
        self.crawled_links = sorted(await self.__links_crawler(url=self.web_instrument.init_url), key=len)
        logger.info(f"Finishing crawling - {self.web_instrument.init_url}")
        return self

    def create_sitemap(self):
        self.file_instrument.create_sitemap(links=self.crawled_links)



import logging.config
from typing import Set, Dict

from .links_crawler import LinksCrawler
from .images_crawler import ImagesCrawler
from .instruments.config import Config

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(module)s.%(funcName)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[
        logging.StreamHandler(),
    ],
)
logger = logging.getLogger(__name__)
__all__ = ("Sitemap",)


class Sitemap:
    def __init__(self, config: Config):
        """
        Main class for work with sitemap images generation

        In this class u can:
            1. Crawling website pages
            2. Generate sitemap images file or get this data
        Args:
            config: dataclass contains all params
        """
        self.config = config

    async def run_images_sitemap(self, url: str) -> None:
        """
        Basic images sitemap generation method
        1. Crawling webpages
        2. Creating images sitemap file
        Args:
            url: website address for crawling
        """
        logger.info(f"Run images sitemap command is started")
        links = await self.crawl_links(url=url)
        await self.generate_images_sitemap_file(links=links)
        logger.info(f"Run images sitemap command finished")

    async def generate_images_sitemap_file(self, links: Set[str]) -> None:
        """
        Method get webpages links set and collect images from them
        And finally generate images sitemap file

        Args:
            links: set with webpages links
        """
        logger.info(f"File generation started")
        images_crawler = ImagesCrawler(config=self.config)
        await images_crawler.create_sitemap(links=links)
        logger.info(f"File generation finished")

    async def images_data(self, links: Set[str]) -> Dict[str, Set[str]]:
        """
        Method collect and return images data as dictionary:
            key - webpage link
            values - set with webpage images
        Args:
            links: pages for parsing

        Returns:
            Dict with collected images data and pages
        """
        images_crawler = ImagesCrawler(config=self.config)
        return await images_crawler.get_data(links=links)

    async def crawl_links(self, url: str) -> Set[str]:
        """
        Method crawling website and collect all domain\subdomain pages
        Args:
            url: website page for starting crawling

        Returns:
            Set of all parsed website pages
        """
        logger.info(f"Pages crawling is started")
        return (await LinksCrawler(init_url=url, config=self.config).run()).crawled_links

    async def run_sitemap(self, url: str) -> None:
        """
        Basic images sitemap generation method
        1. Crawling webpages
        2. Creating images sitemap file
        Args:
            url: website address for crawling
        """
        logger.info(f"Run sitemap command is started")
        (await LinksCrawler(init_url=url, config=self.config).run()).create_sitemap()
        logger.info(f"Run sitemap command finished")



from typing import Dict, List

from .templates import (
    base_image_templ,
    base_loc_template,
    base_url_template,
    base_sitemap_templ,
    base_images_sitemap_templ,
)

__all__ = ("FileInstrument",)


class FileInstrument:
    def __init__(self, file_name: str):
        self.file_name = file_name

    @staticmethod
    def __build_image_sitemap_file(links_images_data: dict[str, List[str]]):
        images_locs = []
        for link, images in links_images_data.items():
            loc = base_loc_template.format(link=link)
            for image_url in images:
                loc += base_image_templ.format(image_url=image_url)
            images_locs.append(base_url_template.format(loc=loc))

        return base_images_sitemap_templ.format(urls_data="".join(images_locs))

    @staticmethod
    def __build_sitemap_file(links: List[str]):
        links_locs = []
        for link in links:
            loc = base_loc_template.format(link=link)
            links_locs.append(base_url_template.format(loc=loc))

        return base_sitemap_templ.format(urls_data="".join(links_locs))

    def __save_file(self, file_data: str):
        with open(self.file_name, "wt") as file:
            file.write(file_data)

    def create_image_sitemap(self, links_images_data: Dict[str, List[str]]):
        file_data = self.__build_image_sitemap_file(links_images_data=links_images_data)
        self.__save_file(file_data=file_data)

    def create_sitemap(self, links: List[str]):
        file_data = self.__build_sitemap_file(links=links)
        self.__save_file(file_data=file_data)



from dataclasses import field, dataclass

__all__ = ("Config",)


@dataclass
class Config:
    """
    accept_subdomains: if True - crawlers will accept subdomains pages/links, else - No
    file_name: sitemap images file name
    """

    max_depth: int = 1
    accept_subdomains: bool = True
    is_query_enabled: bool = True
    file_name: str = "sitemap_images.xml"
    header: dict[str, str] = field(
        default_factory={
            "User-Agent": "ImageSitemap Crawler",
            "Accept": "text/html",
            "Accept-Encoding": "gzip",
            "Connection": "close",
        }
    )



import urllib
import mimetypes
from typing import Set, Dict, List

from .instruments import WebInstrument, FileInstrument
from .instruments.config import Config

__all__ = ("ImagesCrawler",)


class ImagesCrawler:
    def __init__(self, config: Config):
        self.config = config
        if not config.file_name.endswith(".xml"):
            raise ValueError(f"File must be in XML format! Your file name - {self.config.file_name}")
        self.file_instrument = FileInstrument(file_name=self.config.file_name)
        self.web_instrument = WebInstrument

    @staticmethod
    def __filter_images_links(links: Set[str]) -> Set[str]:
        result_links = set()
        for link in links:
            mime_type, _ = mimetypes.guess_type(link if link else "")
            if mime_type and mime_type.startswith("image/") and not link.startswith("data:image/"):
                result_links.add(link)
        return result_links

    async def __parse_images(self, url: str) -> Set[str]:
        links = set()
        if page_data := await self.web_instrument.download_page(url=url):
            images_links = self.__filter_images_links(
                links=self.web_instrument.find_tags(
                    page_data=page_data,
                    tag="img",
                    key="src",
                )
            )
            inner_links = self.web_instrument.filter_inner_links(links=images_links)
            links.update(
                self.web_instrument.filter_links_domain(
                    links=images_links.difference(inner_links), is_subdomain=self.config.accept_subdomains
                )
            )
            links.update({urllib.parse.urljoin(url, inner_link) for inner_link in inner_links})
        return links

    async def __prepare_images_struct(self, links: Set[str]) -> Dict[str, List[str]]:
        images_data = dict()
        all_images = set()

        for url in sorted(links, key=len):
            if parsed_images := (await self.__parse_images(url=url)).difference(all_images):
                images_data.update({url: sorted(parsed_images, key=len)})
                all_images.update(parsed_images)

        return images_data

    async def create_sitemap(self, links: Set[str]):
        self.web_instrument = WebInstrument(init_url=next(iter(links)), config=self.config)
        self.file_instrument.create_image_sitemap(links_images_data=await self.__prepare_images_struct(links=links))

    async def get_data(self, links: Set[str]) -> Dict[str, List[str]]:
        self.web_instrument = WebInstrument(init_url=next(iter(links)), config=self.config)
        return await self.__prepare_images_struct(links=links)



base_images_sitemap_templ = """<?xml version="1.0" encoding="UTF-8"?>
<urlset
\txmlns="http://www.sitemaps.org/schemas/sitemap/0.9"
\txmlns:image="http://www.google.com/schemas/sitemap-image/1.1">
{urls_data}</urlset>\n"""

base_image_templ = """\t\t<image:image>
\t\t\t<image:loc>{image_url}</image:loc>
\t\t</image:image>
"""
base_loc_template = """\t\t<loc>{link}</loc>\n"""
base_url_template = """\t<url>\n{loc}\t</url>\n"""

base_sitemap_templ = """<?xml version="1.0" encoding="UTF-8"?>
<urlset
\txmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
{urls_data}</urlset>\n"""
